import os
import unittest

from pychemstation.control import HPLCController
from pychemstation.utils.sequence_types import *
from tests.constants import *


class TestSequence(unittest.TestCase):
    def setUp(self):
        for path in HEIN_LAB_CONSTANTS:
            if not os.path.exists(path):
                self.fail(
                    f"{path} does not exist on your system. If you would like to run tests, please change this path.")

        self.hplc_controller = HPLCController(data_dir=DATA_DIR,
                                              comm_dir=DEFAULT_COMMAND_PATH,
                                              method_dir=DEFAULT_METHOD_DIR,
                                              sequence_dir=SEQUENCE_DIR)

    def test_switch(self):
        try:
            self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
        except Exception as e:
            self.fail(f"Should have not expected: {e}")

    def test_edit_row(self):
        self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
        self.hplc_controller.edit_sequence_row(SequenceEntry(
                    vial_location=10,
                    method="C:\\ChemStation\\1\\Methods\\General-Poroshell",
                    num_inj=3,
                    inj_vol=4,
                    sample_name="Blank",
                    sample_type=SampleType.BLANK,
                    inj_source=InjectionSource.HIP_ALS
                ), 1)

    def test_edit_entire_table(self):
        self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
        seq_table = SequenceTable(
            name=DEFAULT_SEQUENCE,
            rows=[
                SequenceEntry(
                    vial_location=3,
                    method="C:\\ChemStation\\1\\Methods\\General-Poroshell",
                    num_inj=3,
                    inj_vol=4,
                    sample_name="Control",
                    sample_type=SampleType.CONTROL,
                    inj_source=InjectionSource.MANUAL
                ),
                SequenceEntry(
                    vial_location=1,
                    method="C:\\ChemStation\\1\\Methods\\General-Poroshell",
                    num_inj=1,
                    inj_vol=1,
                    sample_name="Sample",
                    sample_type=SampleType.SAMPLE,
                    inj_source=InjectionSource.AS_METHOD
                ),
                SequenceEntry(
                    vial_location=10,
                    method="C:\\ChemStation\\1\\Methods\\General-Poroshell",
                    num_inj=3,
                    inj_vol=4,
                    sample_name="Blank",
                    sample_type=SampleType.BLANK,
                    inj_source=InjectionSource.HIP_ALS
                ),
            ]
        )
        self.hplc_controller.edit_sequence(seq_table)


    def test_run(self):
        # self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
        seq_table = SequenceTable(
            name=DEFAULT_SEQUENCE,
            rows=[
                SequenceEntry(
                    vial_location=1,
                    method="C:\\ChemStation\\1\\Methods\\" + DEFAULT_METHOD,
                    num_inj=1,
                    inj_vol=1,
                    sample_name="Test",
                    sample_type=SampleType.BLANK
                ),
                SequenceEntry(
                    vial_location=2,
                    method="C:\\ChemStation\\1\\Methods\\" + DEFAULT_METHOD,
                    num_inj=1,
                    inj_vol=1,
                    sample_name="Test2",
                    sample_type=SampleType.BLANK
                ),
            ]
        )
        # self.hplc_controller.edit_sequence(seq_table)
        self.hplc_controller.run_sequence(seq_table)
        chrom = self.hplc_controller.get_last_run_sequence_data()
        self.assertTrue(len(chrom) == 2)



from dataclasses import dataclass
from enum import Enum
from typing import Optional


@dataclass
class TrayLocation:
    row: str
    col: int


@dataclass
class SequenceDataFiles:
    sequence_name: str
    dir: str
    child_dirs: list[str]


class SampleType(Enum):
    SAMPLE = 1
    BLANK = 2
    CALIBRATION = 3
    CONTROL = 4


class InjectionSource(Enum):
    AS_METHOD = "As Method"
    MANUAL = "Manual"
    MSD = "MSD"
    HIP_ALS = "HipAls"


@dataclass
class SequenceEntry:
    sample_name: str
    vial_location: int
    method: Optional[str] = None
    num_inj: Optional[int] = 1
    inj_vol: Optional[int] = 2
    inj_source: Optional[InjectionSource] = InjectionSource.HIP_ALS
    sample_type: Optional[SampleType] = SampleType.SAMPLE


@dataclass
class SequenceTable:
    name: str
    rows: list[SequenceEntry]



from ..utils.table_types import Table

# maximum command number
MAX_CMD_NO = 255

# tables
METHOD_TIMETABLE = Table(
    register="RCPMP1Method[1]",
    name="Timetable"
)

SEQUENCE_TABLE = Table(
    register="_sequence[1]",
    name="SeqTable1"
)



from typing import Any

from copy import deepcopy

import os
import time

from .table_controller import TableController
from ...control import CommunicationController
from ...utils.chromatogram import SEQUENCE_TIME_FORMAT, AgilentHPLCChromatogram
from ...utils.constants import SEQUENCE_TABLE
from ...utils.macro import Command
from ...utils.sequence_types import SequenceTable, SequenceEntry, SequenceDataFiles
from ...utils.table_types import TableOperation, RegisterFlag


class SequenceController(TableController):
    """
    Class containing sequence related logic
    """

    def __init__(self, controller: CommunicationController, src: str, data_dir: str):
        super().__init__(controller, src, data_dir)

    def switch(self, seq_name: str):
        """
        Switch to the specified sequence. The sequence name does not need the '.S' extension.

        :param seq_name: The name of the sequence file
        """
        self.send(f'_SeqFile$ = "{seq_name}.S"')
        self.send(f'_SeqPath$ = "{self.src}"')
        self.send(Command.SWITCH_SEQUENCE_CMD)
        time.sleep(2)
        self.send(Command.GET_SEQUENCE_CMD)
        time.sleep(2)
        # check that method switched
        for _ in range(10):
            try:
                parsed_response = self.receive().splitlines()[1].split()[1:][0]
                break
            except IndexError:
                continue

        assert parsed_response == f"{seq_name}.S", "Switching sequence failed."

    def edit(self, sequence_table: SequenceTable):
        """
        Updates the currently loaded sequence table with the provided table. This method will delete the existing sequence table and remake it.
        If you would only like to edit a single row of a sequence table, use `edit_sequence_table_row` instead.

        :param sequence_table:
        """
        self.send("Local Rows")
        self.sleep(1)
        self.delete_table(SEQUENCE_TABLE)
        self.sleep(1)
        self.new_table(SEQUENCE_TABLE)
        self.sleep(1)
        self._get_table_rows(SEQUENCE_TABLE)

        for _ in sequence_table.rows:
            self.add_table_row(SEQUENCE_TABLE)
            self.sleep(1)
            self.send(Command.SAVE_SEQUENCE_CMD)
        self._get_table_rows(SEQUENCE_TABLE)
        self.send(Command.SAVE_SEQUENCE_CMD)
        self.send(Command.SWITCH_SEQUENCE_CMD)

        for i, row in enumerate(sequence_table.rows):
            self.edit_row(row=row, row_num=i + 1)
            self.sleep(1)
        self.send(Command.SAVE_SEQUENCE_CMD)


    def edit_row(self, row: SequenceEntry, row_num: int):
        """
        Edits a row in the sequence table. Assumes the row already exists.

        :param row: sequence row entry with updated information
        :param row_num: the row to edit, based on -1-based indexing
        """
        if row.vial_location:
            self.sleepy_send(TableOperation.EDIT_ROW_VAL.value.format(register=SEQUENCE_TABLE.register,
                                                                      table_name=SEQUENCE_TABLE.name,
                                                                      row=row_num,
                                                                      col_name=RegisterFlag.VIAL_LOCATION,
                                                                      val=row.vial_location))
        if row.method:
            self.sleepy_send(TableOperation.EDIT_ROW_TEXT.value.format(register=SEQUENCE_TABLE.register,
                                                                       table_name=SEQUENCE_TABLE.name,
                                                                       row=row_num,
                                                                       col_name=RegisterFlag.METHOD,
                                                                       val=row.method))

        if row.num_inj:
            self.sleepy_send(TableOperation.EDIT_ROW_VAL.value.format(register=SEQUENCE_TABLE.register,
                                                                      table_name=SEQUENCE_TABLE.name,
                                                                      row=row_num,
                                                                      col_name=RegisterFlag.NUM_INJ,
                                                                      val=row.num_inj))

        if row.inj_vol:
            self.sleepy_send(TableOperation.EDIT_ROW_TEXT.value.format(register=SEQUENCE_TABLE.register,
                                                                       table_name=SEQUENCE_TABLE.name,
                                                                       row=row_num,
                                                                       col_name=RegisterFlag.INJ_VOL,
                                                                       val=row.inj_vol))

        if row.inj_source:
            self.sleepy_send(TableOperation.EDIT_ROW_TEXT.value.format(register=SEQUENCE_TABLE.register,
                                                                       table_name=SEQUENCE_TABLE.name,
                                                                       row=row_num,
                                                                       col_name=RegisterFlag.INJ_SOR,
                                                                       val=row.inj_source.value))

        if row.sample_name:
            self.sleepy_send(TableOperation.EDIT_ROW_TEXT.value.format(register=SEQUENCE_TABLE.register,
                                                                       table_name=SEQUENCE_TABLE.name,
                                                                       row=row_num,
                                                                       col_name=RegisterFlag.NAME,
                                                                       val=row.sample_name))
            self.sleepy_send(TableOperation.EDIT_ROW_TEXT.value.format(register=SEQUENCE_TABLE.register,
                                                                       table_name=SEQUENCE_TABLE.name,
                                                                       row=row_num,
                                                                       col_name=RegisterFlag.DATA_FILE,
                                                                       val=row.sample_name))
        if row.sample_type:
            self.sleepy_send(TableOperation.EDIT_ROW_VAL.value.format(register=SEQUENCE_TABLE.register,
                                                                      table_name=SEQUENCE_TABLE.name,
                                                                      row=row_num,
                                                                      col_name=RegisterFlag.SAMPLE_TYPE,
                                                                      val=row.sample_type.value))

    def run(self, sequence_table: SequenceTable):
        """
        Starts the currently loaded sequence, storing data
        under the <data_dir>/<sequence table name> folder.
        Device must be ready.

        :param sequence_table:
        """
        timestamp = time.strftime(SEQUENCE_TIME_FORMAT)
        # self.send(Command.RUN_SEQUENCE_CMD.value)
        folder_name = f"{sequence_table.name} {timestamp}"
        subdirs = [x[0] for x in os.walk(self.data_dir)]
        # time.sleep(60)
        potential_folders = sorted(list(filter(lambda d: folder_name in d, subdirs)))
        parent_folder = potential_folders[0]
        self.data_files.append(SequenceDataFiles(
            sequence_name=sequence_table.name,
            dir=parent_folder,
            child_dirs=[r.sample_name + ".D" for r in sequence_table.rows]))

    def retrieve_recent_data_files(self):
        sequence_data_files: SequenceDataFiles = self.data_files[-1]
        return os.path.join(sequence_data_files.dir, sequence_data_files.child_dirs[-1])

    def get_data(self) -> tuple[bool, Any]:
        data_ready = True # self.data_ready()
        sequence_data_files: SequenceDataFiles = self.data_files[-1]
        spectra: list[dict[str, AgilentHPLCChromatogram]] = []
        if data_ready:
            for row in sequence_data_files.child_dirs:
                data_path = os.path.join(sequence_data_files.dir, row)
                self.get_spectrum(data_path)
                spectra.append(deepcopy(self.spectra))
            return data_ready, spectra
        else:
            return False, None




# CONSTANTS: paths only work in Hein group HPLC machine in room 242
DEFAULT_COMMAND_PATH = "C:\\Users\\User\\Desktop\\Lucy\\"
DEFAULT_METHOD = "GENERAL-POROSHELL-OPT"
DEFAULT_TESTING_METHOD = "GENERAL-POROSHELL-MIN"
DEFAULT_METHOD_DIR = "C:\\ChemStation\\1\\Methods\\"
DATA_DIR = "C:\\Users\\Public\\Documents\\ChemStation\\3\\Data"
SEQUENCE_DIR = "C:\\USERS\\PUBLIC\\DOCUMENTS\\CHEMSTATION\\3\\Sequence"
DEFAULT_SEQUENCE = "hplc_testing"
METHOD_EXAMPLE = "/Users/lucyhao/Codes/heinanalyticalcontrol/tests/methods/General-Poroshell.M"

HEIN_LAB_CONSTANTS = [DEFAULT_COMMAND_PATH, DEFAULT_METHOD_DIR, DATA_DIR, SEQUENCE_DIR]



"""
Abstract module containing shared logic for Method and Sequence tables.

Authors: Lucy Hao
"""

import abc
import os
from typing import Union, Optional

import polling

from ...control import CommunicationController
from ...utils.chromatogram import AgilentHPLCChromatogram
from ...utils.macro import Command
from ...utils.method_types import MethodTimetable
from ...utils.sequence_types import SequenceDataFiles
from ...utils.table_types import Table, TableOperation, RegisterFlag


class TableController(abc.ABC):

    def __init__(self, controller: CommunicationController, src: str, data_dir: str):
        self.controller = controller
        if os.path.isdir(src):
            self.src: str = src
        else:
            raise FileNotFoundError(f"dir: {src} not found.")

        if os.path.isdir(data_dir):
            self.data_dir: str = data_dir
        else:
            raise FileNotFoundError(f"dir: {data_dir} not found.")

        self.spectra = {
            "A": AgilentHPLCChromatogram(self.data_dir),
            "B": AgilentHPLCChromatogram(self.data_dir),
            "C": AgilentHPLCChromatogram(self.data_dir),
            "D": AgilentHPLCChromatogram(self.data_dir),
        }

        self.data_files: Union[list[SequenceDataFiles], list[str]] = []

    def receive(self):
        return self.controller.receive()

    def send(self, cmd: Union[Command, str]):
        self.controller.send(cmd)

    def sleepy_send(self, cmd: Union[Command, str]):
        self.controller.sleepy_send(cmd)

    def sleep(self, seconds: int):
        """
        Tells the HPLC to wait for a specified number of seconds.

        :param seconds: number of seconds to wait
        """
        self.send(Command.SLEEP_CMD.value.format(seconds=seconds))

    def add_table_row(self, table: Table):
        """
        Adds a row to the provided table for currently loaded method or sequence.
        Import either the SEQUENCE_TABLE or METHOD_TIMETABLE from hein_analytical_control.constants.
        You can also provide your own table.

        :param table: the table to add a new row to
        """
        self.sleepy_send(TableOperation.NEW_ROW.value.format(register=table.register,
                                                             table_name=table.name))

    def delete_table(self, table: Table):
        """
        Deletes the table for the current loaded method or sequence.
        Import either the SEQUENCE_TABLE or METHOD_TIMETABLE from hein_analytical_control.constants.
        You can also provide your own table.

        :param table: the table to delete
        """
        self.sleepy_send(TableOperation.DELETE_TABLE.value.format(register=table.register,
                                                                  table_name=table.name))

    def new_table(self, table: Table):
        """
        Creates the table for the currently loaded method or sequence. Import either the SEQUENCE_TABLE or
        METHOD_TIMETABLE from hein_analytical_control.constants. You can also provide your own table.

        :param table: the table to create
        """
        self.send(TableOperation.CREATE_TABLE.value.format(register=table.register,
                                                           table_name=table.name))

    def _get_table_rows(self, table: Table) -> str:
        self.send(TableOperation.GET_OBJ_HDR_VAL.value.format(internal_val="Rows",
                                                              register=table.register,
                                                              table_name=table.name,
                                                              col_name=RegisterFlag.NUM_ROWS, ))
        res = self.controller.receive()
        self.send("Sleep 1")
        self.send('Print Rows')
        return res

    def check_hplc_ready_with_data(self, method: Optional[MethodTimetable] = None) -> bool:
        """
        Checks if ChemStation has finished writing data and can be read back.

        :param method: if you are running a method and want to read back data, the timeout period will be adjusted to be longer than the method's runtime
        :return: Return True if data can be read back, else False.
        """
        self.controller.set_status()

        timeout = 10 * 60
        hplc_run_done = polling.poll(
            lambda: self.controller.check_data(self.retrieve_recent_data_files()),
            timeout=timeout,
            step=30
        )

        return hplc_run_done

    @abc.abstractmethod
    def retrieve_recent_data_files(self):
        pass

    @abc.abstractmethod
    def get_data(self) -> tuple[bool,]:
        pass

    def get_spectrum(self, data_file: str):
        """
        Load chromatogram for any channel in spectra dictionary.
        """
        for channel, spec in self.spectra.items():
            spec.load_spectrum(data_path=data_file, channel=channel)

    def data_ready(self) -> bool:
        return self.check_hplc_ready_with_data()



import os

import unittest

from pychemstation.control import HPLCController
from pychemstation.utils.macro import *
from tests.constants import *


class TestComm(unittest.TestCase):

    def setUp(self):
        for path in HEIN_LAB_CONSTANTS:
            if not os.path.exists(path):
                self.fail(
                    f"{path} does not exist on your system. If you would like to run tests, please change this path.")

        self.hplc_controller = HPLCController(data_dir=DATA_DIR,
                                              comm_dir=DEFAULT_COMMAND_PATH,
                                              method_dir=DEFAULT_METHOD_DIR,
                                              sequence_dir=SEQUENCE_DIR)

    def test_status_check_standby(self):
        self.hplc_controller.standby()
        self.assertTrue(self.hplc_controller.status()[0] in [HPLCAvailStatus.STANDBY, HPLCRunningStatus.NOTREADY])

    def test_status_check_preprun(self):
        self.hplc_controller.preprun()
        self.assertTrue(self.hplc_controller.status()[0] in [HPLCAvailStatus.PRERUN, HPLCAvailStatus.STANDBY,
                                                             HPLCRunningStatus.NOTREADY])

    def test_send_command(self):
        try:
            self.hplc_controller.send(Command.GET_METHOD_CMD)
        except Exception as e:
            self.fail(f"Should not throw error: {e}")

    def test_send_str(self):
        try:
            self.hplc_controller.send("Local TestNum")
            self.hplc_controller.send("TestNum = 0")
        except Exception as e:
            self.fail(f"Should not throw error: {e}")

    def test_get_response(self):
        try:
            self.hplc_controller.switch_method(method_name=DEFAULT_METHOD)
            self.hplc_controller.send(Command.GET_METHOD_CMD)
            res = self.hplc_controller.receive()
            self.assertTrue(DEFAULT_METHOD in res)
        except Exception as e:
            self.fail(f"Should not throw error: {e}")

    def test_pump_lamp(self):
        pump_lamp = [
            ("response", self.hplc_controller.lamp_on),
            ("response", self.hplc_controller.lamp_off),
            ("response", self.hplc_controller.pump_on),
            ("response", self.hplc_controller.pump_off),
        ]

        for operation in pump_lamp:
            try:
                operation[1]()
            except Exception as e:
                self.fail(f"Failed due to: {e}")



import os
import unittest

from pychemstation.control import HPLCController
from pychemstation.utils.method_types import *
from tests.constants import *


class TestMethod(unittest.TestCase):
    def setUp(self):
        for path in HEIN_LAB_CONSTANTS:
            if not os.path.exists(path):
                self.fail(
                    f"{path} does not exist on your system. If you would like to run tests, please change this path.")

        self.hplc_controller = HPLCController(data_dir=DATA_DIR,
                                              comm_dir=DEFAULT_COMMAND_PATH,
                                              method_dir=DEFAULT_METHOD_DIR,
                                              sequence_dir=SEQUENCE_DIR)



    def test_load_method_details(self):
        self.hplc_controller.switch_method(DEFAULT_METHOD)
        try:
            gp_mtd = self.hplc_controller.method_controller.load(DEFAULT_METHOD)
            self.assertTrue(gp_mtd.first_row.organic_modifier.val == 5)
        except Exception as e:
            self.fail(f"Should have not failed, {e}")

    def test_method_update_timetable(self):
        self.hplc_controller.method_controller.switch(DEFAULT_METHOD)
        new_method = MethodTimetable(
            first_row=HPLCMethodParams(
                organic_modifier=7,
                flow=0.44,
                maximum_run_time=2,
                temperature=15),
            subsequent_rows=[
                TimeTableEntry(
                    start_time=0.10,
                    organic_modifer=7,
                    flow=0.34),
                TimeTableEntry(
                    start_time=1,
                    organic_modifer=99,
                    flow=0.55)])
        try:
            self.hplc_controller.edit_method(new_method)
        except Exception as e:
            self.fail(f"Should have not failed: {e}")

    def test_run_method(self):
        try:
            self.hplc_controller.run_method(experiment_name="test_experiment")
            chrom = self.hplc_controller.get_last_run_method_data()
        except Exception as e:
            self.fail(f"Should have not failed: {e}")




if __name__ == '__main__':
    unittest.main()



"""
Module to provide API for higher-level HPLC actions.

Authors: Lucy Hao
"""

from typing import Union, Optional

from ..control import CommunicationController
from ..control.table import MethodController, SequenceController
from ..utils.chromatogram import AgilentHPLCChromatogram
from ..utils.macro import Command, HPLCRunningStatus, HPLCAvailStatus, HPLCErrorStatus
from ..utils.method_types import MethodTimetable
from ..utils.sequence_types import SequenceTable, SequenceEntry


class HPLCController:
    def __init__(self,
                 comm_dir: str,
                 data_dir: str,
                 method_dir: str,
                 sequence_dir: str):
        """Initialize HPLC controller. The `hplc_talk.mac` macro file must be loaded in the Chemstation software.
        `comm_dir` must match the file path in the macro file.

        :param comm_dir: Name of directory for communication, where ChemStation will read and write from. Can be any existing directory.
        :raises FileNotFoundError: If either `data_dir`, `method_dir` or `comm_dir` is not a valid directory.
        """
        self.comm = CommunicationController(comm_dir=comm_dir)
        self.method_controller = MethodController(controller=self.comm, src=method_dir, data_dir=data_dir)
        self.sequence_controller = SequenceController(controller=self.comm, src=sequence_dir, data_dir=data_dir)

    def send(self, cmd: Union[Command, str]):
        self.comm.send(cmd)

    def receive(self) -> str:
        return self.comm.receive()

    def status(self) -> list[HPLCRunningStatus | HPLCAvailStatus | HPLCErrorStatus]:
        return self.comm.get_status()

    def switch_method(self, method_name: str):
        """
        Allows the user to switch between pre-programmed methods. No need to append '.M'
        to the end of the method name. For example. for the method named 'General-Poroshell.M',
        only 'General-Poroshell' is needed.

        :param method_name: any available method in Chemstation method directory
        :raise IndexError: Response did not have expected format. Try again.
        :raise AssertionError: The desired method is not selected. Try again.
        """
        self.method_controller.switch(method_name)

    def switch_sequence(self, sequence_name: str):
        """
         Allows the user to switch between pre-programmed sequences. The sequence name does not need the '.S' extension.
         For example. for the method named 'mySeq.S', only 'mySeq' is needed.

        :param seq_name: The name of the sequence file
        """
        self.sequence_controller.switch(sequence_name)

    def run_method(self, experiment_name: str):
        """
        This is the preferred method to trigger a run.
        Starts the currently selected method, storing data
        under the <data_dir>/<experiment_name>.D folder.
        The <experiment_name> will be appended with a timestamp in the '%Y-%m-%d-%H-%M' format.
        Device must be ready.

        :param experiment_name: Name of the experiment
        """
        self.method_controller.run(experiment_name)

    def run_sequence(self, sequence_table: SequenceTable):
        """
        Starts the currently loaded sequence, storing data
        under the <data_dir>/<sequence table name> folder.
        Device must be ready.

        :param sequence_table:
        """
        self.sequence_controller.run(sequence_table)

    def edit_method(self, updated_method: MethodTimetable):
        """Updated the currently loaded method in ChemStation with provided values.

        :param updated_method: the method with updated values, to be sent to Chemstation to modify the currently loaded method.
        """
        self.method_controller.edit(updated_method)

    def edit_sequence(self, updated_sequence: SequenceTable):
        """
        Updates the currently loaded sequence table with the provided table. This method will delete the existing sequence table and remake it.
        If you would only like to edit a single row of a sequence table, use `edit_sequence_row` instead.

        :param sequence_table:
        """
        self.sequence_controller.edit(updated_sequence)

    def edit_sequence_row(self, row: SequenceEntry, num: int):
        """
        Edits a row in the sequence table. Assumes the row already exists.

        :param row: sequence row entry with updated information
        :param row_num: the row to edit, based on -1-based indexing
        """
        self.sequence_controller.edit_row(row, num)

    def get_last_run_method_data(self) -> Optional[dict[str, AgilentHPLCChromatogram]]:
        """
        Returns the last run method data.
        """
        data_valid, data = self.method_controller.get_data()
        if data_valid:
            return data
        return None

    def get_last_run_sequence_data(self) -> Optional[list[dict[str, AgilentHPLCChromatogram]]]:
        """
        Returns data for all rows in the last run sequence data.
        """
        data_valid, data = self.sequence_controller.get_data()
        if data_valid:
            return data
        return None

    def standby(self):
        """Switches all modules in standby mode. All lamps and pumps are switched off."""
        self.send(Command.STANDBY_CMD)

    def preprun(self):
        """ Prepares all modules for run. All lamps and pumps are switched on."""
        self.send(Command.PREPRUN_CMD)

    def lamp_on(self):
        """Turns the UV lamp on."""
        self.send(Command.LAMP_ON_CMD)

    def lamp_off(self):
        """Turns the UV lamp off."""
        self.send(Command.LAMP_OFF_CMD)

    def pump_on(self):
        """Turns on the pump on."""
        self.send(Command.PUMP_ON_CMD)

    def pump_off(self):
        """Turns the pump off."""
        self.send(Command.PUMP_OFF_CMD)
